{% extends 'Agent/main.jinja2' %}

{% block context %}
Available visualization types:
{{ doc }}

The current SQL table name: {{ memory['pipeline'].table }}.

Here is the current dataset to use:
{{ memory['data'] }}

{%- if "view" in memory and memory["view"].get("agent") and actor_name.startswith(memory["view"].get("agent")) %}
Adapt from the previous view specification below as needed:
```yaml
{{ memory["view"]["spec"] | json_to_yaml }}
```
{%- endif %}

{% if memory['pipeline'].source.dialect == "snowflake" %}
Please use all upper case for field names.
{% endif %}
{%- endblock %}
